// Unit.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;
import java.io.*;
/**
 * The Unit class instances are micro unit in our virtual enterprise;
 * i.e. the units where the steps required to build a product (to fulfill
 * an order) are done
 *
 * @author Pietro Terna (with Marco Remondino, Dario Landini, Nicola 
 * Ormezzano and Paolo Pelligra contributions)
 */
public class Unit extends SwarmObjectImpl implements PTHistogramPlottable{

    /** the number identifying the unit*/
    public int unitNumber;
    /** the unit procurementAssembler */
    ProcurementAssembler myProcurementAssembler;
    /** the unit standAloneBatchAssembler */
    StandAloneBatchAssembler myStandAloneBatchAssembler;
    /** the unit sequentialBatchAssembler */
    SequentialBatchAssembler mySequentialBatchAssembler;
    /** the number of different production phases the unit is able to 
     * deal with (the 0 copy of this field is used for technical reasons) */
    public int nOfPhasesToDealWith=1, countPhases=0;
    /** dimension or the vector unitProductionPhases for a 
     *  simple unit or redimension for a complex one */
    public boolean nOfPhasesToDealWithDimensionOrRedimension=true;
    /** the unit production phase */
    public int[] unitProductionPhases;

    /** the current state of production phase (what are we doing) 
     *  
     *  the current set is established by the order currently produced 
     *  by this unit in this tick
     * */
    public int currentState=-1; // -1 stays for undefined state
    public int temporaryCurrentState;

    /** choosing - global choice - if we use or not the warehouses
	(to provide or not inventories of our products) */
    public boolean useWarehouses;
    /** choosing if we use or not newses about incoming productions*/
    public boolean useNewses;
    /** a flag about using pieces in warehouse */
    public boolean madeProduction;
    /** the list of the normal units existing in our environment*/
    public ListImpl unitList;
    /** the list of end units existing in our environment*/
    public ListImpl endUnitList;
    /** the list of the orders to be executed */
    public ListImpl  waitingList;
    /** its iterator */
    public ListIndex waitingListIndex;
    /** the list of the orders in the waitingList, not kept into
     *  any parallel unit, such as a myProcurementAssembler */
    public ListImpl usableWaitingList;
    /** the list of the orders in the waitingList, kept into
     *  myStandAloneBatchAssembler */
    public ListImpl standAloneBatchWaitingList;
    /** the list of the orders in the waitingList to be produced as a 
     *  sequential batch */
    public ListImpl sequentialBatchWaitingList;
    /** its iterator */
    public ListIndex sequentialBatchWaitingListIndex;
    /** the list containing the orders (if any) performed in a day, also
	via the inventories contained in our warehouse */
    public ListImpl dailyProductionList;
    /** the modelSwarm (to which we apply the 
     * setProductionTimeAndRecipeLength(...) method */
    VEFrameModelSwarm vEFrameModelSwarm;
    /** the temporary address of an existing unit */
    public Unit aUnit;
    /** the temporary address of an existing endUnit */
    public EndUnit anEndUnit;
     /** the temporary address of a unit to wich send a news */
    public Unit receivingUnit;
    /** our warehouse, if any */
    public Warehouse myWarehouse;
    /** the first order to be executed */
    public Order firstOrder;
    /** an order inserted in a news */
    public Order incomingOrder;
    /** a news to be sent */
    public News aNews;
    /** a news to be deleted a the arrive of the related order */
    public News toBeDroppedNews;
    /** our rule master about inventories */
    public InventoryRuleMaster inventoryRuleMaster;
     /** our rule master about information system */
    public InformationRuleMaster informationRuleMaster;
    /** the list of the newses */
    public ListImpl newsList;
    /** how deep we look at production steps to propagate news (how many steps 
     * beyond the current production phase)*/
    public int informationDeepness;
    /** of order for which newses are propagated */
    public int nOfOrdersInNewses;
    /** the whole list of the orders */
    public ListImpl wholeOrderList;

    /** the number of different values of fixed and variable costs,
     *  related to the number of production phases */
    public int nOfCostTypes=1, countCostTypes=0;
    /** dimension or the vectors fixedCosts and variableCosts for a 
     *  simple unit or redimension for a complex one */
    public boolean nOfCostTypesDimensionOrRedimension=true;
    /** our fixed and variable costs */
    public float[] fixedCosts, variableCosts;

    /** the number of phase inventory choices,
     *  related to the number of production phases */
    public int nOfPhaseInventories=1, countPhaseInventories=0; 
    /** dimension or the vector phaseInventories for a 
     *  simple unit or redimension for a complex one */
    public boolean nOfPhaseInventoriesDimensionOrRedimension=true;
    /** our phase inventory choices */
    public boolean[] phaseInventories;

    /** total tick variable costs */
    public float totalTickVariableCosts;
    /** total daily fixed and variable costs */
    public float totalDailyCosts;
    /** the costs from the beginning of the simulation */
    public float totalCosts = 0;
    /** the daily revenues from the finished orders (sold) */
    public float dailySaleRevenues;
    /** the revenues from the beginning of the simulation */
    public float totalSaleRevenues = 0;
    /** the costs of the finished orders */
    public float totalFinishedOrderCosts = 0;
    /** the daily evaluation of the stored component */
    public float dailyStoredComponentValue = 0;
    /** the cumulative inventory financial costs generated from inventories */
    float totalInventoryFinancialCosts = 0;
    /** the writer dedicated to the log of the concluded orders */
    MyWriter concludedOrderLog;
    /** the instance of AssigningTool we use */
    AssigningTool assigningTool; 

    /**
     * the first constructor for Unit (using warehouses)
     */
    public Unit (Zone aZone, ListImpl ul, ListImpl eul, VEFrameModelSwarm mo,
		 InventoryRuleMaster rm, Warehouse wh, ListImpl ol,
		 ProcurementAssembler pa, StandAloneBatchAssembler ba,
		 SequentialBatchAssembler sa, MyWriter col, AssigningTool at)
    {
	// Call the constructor for the Unit parent class.
	super(aZone);

	useWarehouses = true;            // if we are using this constructor,
	                                 // warehouses are activated
	useNewses = false;               // if we are using this constructor,
	                                 // newses are not activated
        inventoryRuleMaster = rm;        // our InventoryRuleMaster
	myWarehouse = wh;	         // our warehouse address
	myProcurementAssembler = pa;     // our procurement assembler
	myStandAloneBatchAssembler = ba; // our stand alone batch assembler 
	mySequentialBatchAssembler = sa; // our sequential  batch assembler 

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// its iterator
	waitingListIndex = waitingList.listBegin(getZone());
	// the the list of the orders in the waitingList, not kept into
	//  any parallel unit, such as a myProcurementAssembler
	usableWaitingList = new ListImpl (getZone());
	// the list of the orders in the waitingList, kept into
	//  myStandAloneBatchAssembler 
	standAloneBatchWaitingList = new ListImpl (getZone());
	//  the list of the orders in the waitingList to be produced as a 
	// sequential batch 
	sequentialBatchWaitingList = new ListImpl (getZone());
	// its iterator
	sequentialBatchWaitingListIndex = sequentialBatchWaitingList.
	    listBegin(getZone());
	// the list of the normal units
	unitList=ul;
	// the list of the end units
	endUnitList=eul;
	// its iterator
	vEFrameModelSwarm = mo;
	// the whole list of orders
        wholeOrderList = ol;
	// the daily list of finished orders
	dailyProductionList= new ListImpl(getZone());
	// the log writer
	concludedOrderLog=col;
	// the assigning tool
	assigningTool=at;

    }

    /**
     * the second constructor for Unit (without warehouses)
     */
    public Unit (Zone aZone, ListImpl ul, ListImpl eul, VEFrameModelSwarm mo,
                  ListImpl ol, ProcurementAssembler pa, 
		 StandAloneBatchAssembler ba,
		 SequentialBatchAssembler sa, MyWriter col, AssigningTool at) {

	// Call the constructor for the Unit parent class
	super(aZone);

	useWarehouses = false;           // if we are using this constructor,
	                                 // warehouses are not activated
	useNewses = false;               // if we are using this constructor,
	                                 // newses are not activated
        inventoryRuleMaster = null;      // to avoid compiler warning/error
	myWarehouse = null;              // " " "
	myProcurementAssembler = pa;     // our procurement assembler
	myStandAloneBatchAssembler = ba; // our stand alone batch assembler 
	mySequentialBatchAssembler = sa; // our sequential  batch assembler 

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// its iterator
	waitingListIndex = waitingList.listBegin(getZone());
	// the the list of the orders in the waitingList, not kept into
	//  any parallel unit, such as a procurementAssembler
	usableWaitingList = new ListImpl (getZone());
	// the list of the orders in the waitingList, kept into
	//  myStandAloneBatchAssembler 
	standAloneBatchWaitingList = new ListImpl (getZone());
	//  the list of the orders in the waitingList to be produced as a 
	// sequential batch 
	sequentialBatchWaitingList = new ListImpl (getZone());
	// its iterator
	sequentialBatchWaitingListIndex = sequentialBatchWaitingList.
	    listBegin(getZone());
	// the list of the normal units
	unitList=ul;
	// the list of the end units
	endUnitList=eul;
	// the model
	vEFrameModelSwarm = mo;
	// the daily list of finished orders
	dailyProductionList= new ListImpl(getZone());
	// the whole list of orders
        wholeOrderList = ol;
	// the log writer
	concludedOrderLog=col;
	// the assigning tool
	assigningTool=at;

   }

    /**
     * the third constructor for Unit (with warehouses and newses)
     */
    public Unit (Zone aZone, ListImpl ul, ListImpl eul, VEFrameModelSwarm mo,
		 InventoryRuleMaster rm, Warehouse wh, 
		 InformationRuleMaster irm, int idp, int noin, ListImpl ol,
		 ProcurementAssembler pa, StandAloneBatchAssembler ba,
		 SequentialBatchAssembler sa, MyWriter col, AssigningTool at) 
    {
	// Call the constructor for the Unit parent class.
	super(aZone);

	useWarehouses = true;            // if we are using this constructor,
	                                 // warehouses are activated
        useNewses = true;                // if we are using this constructor,
	                                 // newses are activated
	inventoryRuleMaster = rm;        // our InventoryRuleMaster
	informationRuleMaster= irm;      // our InformationRuleMaster
        myWarehouse = wh;	         // our warehouse address
	myProcurementAssembler = pa;     // our procurement assembler
	myStandAloneBatchAssembler = ba; // our stand alone batch assembler 
	mySequentialBatchAssembler = sa; // our sequential  batch assembler 

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// its iterator
	waitingListIndex = waitingList.listBegin(getZone());
	// the the list of the orders in the waitingList, not kept into
	//  any parallel unit, such as a procurementAssembler
	usableWaitingList = new ListImpl (getZone());
	// the list of the orders in the waitingList, kept into
	//  myStandAloneBatchAssembler 
	standAloneBatchWaitingList = new ListImpl (getZone());
	//  the list of the orders in the waitingList to be produced as a 
	// sequential batch 
	sequentialBatchWaitingList = new ListImpl (getZone());
	// its iterator
	sequentialBatchWaitingListIndex = sequentialBatchWaitingList.
	    listBegin(getZone());
	// the list of the normal units
	unitList=ul;
	// the list of the end units
	endUnitList=eul;
	// the model
	vEFrameModelSwarm = mo;
	// the daily list if finished orders
	dailyProductionList= new ListImpl(getZone());
        // the news list
        newsList= new ListImpl(getZone());
        // how many steps we consider in recipes to propagate newses
        informationDeepness= idp;
	// # of orders for which newses are propagated
	nOfOrdersInNewses = noin;    
	// the whole list of orders
        wholeOrderList = ol;
	// the log writer
	concludedOrderLog=col;
	// the assigning tool
	assigningTool=at;

   }

    /** set unit number */
    public void setUnitNumber(int un)
    {
	// the unit id number.
	unitNumber=un;
	// announce the unit to the console
	if(StartVEFrame.verbose)
	System.out.println("Unit number " + unitNumber + " has been created.");
     }

    /** get unit number */
    public int getUnitNumber()
    {
	// the unit id number.
	return unitNumber;
    }

    /** zeroing daily costs */
    public void unitStep0 ()
    {
	// at the beginnig of the day we have not yet generated any cost or
	// revenue 
        totalDailyCosts = 0;
	dailySaleRevenues = 0;

	// choosing an undefined state
	currentState=-1;
    }

    /** copying the orders in the waitingList, if they are not kept into
     *  any parallel unit, such as a myProcurementAssembler or 
     *  mySequentialBatchAssembler */
    public void setUsableWaitingList(){
	int i;
	Order tmp;

	// voiding usableWaitingList
	while(usableWaitingList.getCount()>0)usableWaitingList.removeFirst();

	if(waitingList.getCount()==0)return;
	waitingListIndex.setOffset(0);
	for(i=0;i<waitingList.getCount();i++)
	    {
		tmp=(Order) waitingListIndex.get();
		if(! (myProcurementAssembler.thisOrderIsInTheWaitingList(tmp)
		      || 
		      mySequentialBatchAssembler.thisOrderIsInTheWaitingList(
								     tmp) 

		      || tmp.getInSequentialBatch()  ) )
		    usableWaitingList.addLast(tmp);
		waitingListIndex.next();
	    }

    }

    /** copying the orders in standAloneBatchWaitingList, if they are
     *  kept int a parallel unit, such as a myStandAloneAssembler */
    public void setStandAloneBatchWaitingList(){
	int i;
	Order tmp;

	// voiding standAloneBatchWaitingList
	while(standAloneBatchWaitingList.getCount()>0)
	    standAloneBatchWaitingList.removeFirst();

	if(waitingList.getCount()==0)return;

	waitingListIndex.setOffset(0);
	for(i=0;i<waitingList.getCount();i++)
	    {
		tmp=(Order) waitingListIndex.get();
		if(myStandAloneBatchAssembler.thisOrderIsInTheWaitingList(tmp))
		    standAloneBatchWaitingList.addLast(tmp);
		waitingListIndex.next();
	    }

    }

    /** copying the orders in sequentialBatchWaitingList, if the first order
     *  in the waiting list is in a sequential batch */
    public void setSequentialBatchWaitingList(){
	int i, j, count;

	if (waitingList.getCount()==0)return;

	// if the sequential batch waiting list is not empty we cannot
	// deal with a new batch (anyway, those order are non placed in the
	// usable waiting list, look at setUsableWaitingList
	if(sequentialBatchWaitingList.getCount()>0)return;

	if(! ((Order) waitingList.atOffset(0)).getInSequentialBatch() )return;

	count=((SequentialBatchSpecificationSet) 
	       ((Order) waitingList.atOffset(0))
	       .getPendingSequentialBatchSpecificationSet())
	    .getQuantityInSequentialBatch();

	for (i=0;i<count;i++)
	    {
		if(waitingList.getCount()==0 || 
		   ! ((Order) waitingList.atOffset(0)).getInSequentialBatch() )
		    {
			System.out.println(
			 "Internal error using a sequential batch in unit "+
			 unitNumber);
			System.exit(1);
		    }

		sequentialBatchWaitingList.addLast(waitingList.removeFirst());

	    }

	if(StartVEFrame.verbose)
	    {
		System.out.print("time"+Globals.env.getCurrentTime()+
				 "Length of the seq.w.l. "
				 +sequentialBatchWaitingList.getCount()+
				 " in unit "+unitNumber+" containg orders ");
		for (j=0;j<sequentialBatchWaitingList.getCount();j++)
		    System.out.print(((Order) 
				      sequentialBatchWaitingList.atOffset(j)).
				     getOrderNumber()+" ");
		System.out.println(" ");
	    }

    }

    /** 
     *  managing off line the sequential batches (not considering warehouses)
     */
    public void	manageSequentialBatchWaitingList(){
	int residualTime, quantityInSequentialBatch, i;
	Order ourFirstOrder;
	SequentialBatchSpecificationSet ourSequentialBatchSpecificationSet;

	ourFirstOrder = (Order) sequentialBatchWaitingList.atOffset(0);
	ourSequentialBatchSpecificationSet = (SequentialBatchSpecificationSet)
	    ourFirstOrder.getPendingSequentialBatchSpecificationSet();

	residualTime = ourSequentialBatchSpecificationSet
	    .getProductionTimeInTicks();

	quantityInSequentialBatch = ourSequentialBatchSpecificationSet
	    .getQuantityInSequentialBatch();

	currentState=-1;
	for (i=0; i<getNOfPhasesToDealWith() && currentState==-1;i++)
	    if(ourFirstOrder.getNextStep()==getProductionPhase(i))
		currentState=i;
	if(currentState==-1){
	    System.out.println("Unrecoverable error in unit " + unitNumber);
	    System.exit(1);
	}

	addTickVariableCosts();
	
	sequentialBatchWaitingListIndex.setOffset(0);
	for(i=0;i<sequentialBatchWaitingList.getCount();i++)
	    {
		((Order) sequentialBatchWaitingListIndex.get())
		    .setFixedAndVariableOrderCosts(
						   fixedCosts[currentState]/
						   quantityInSequentialBatch, 
						   variableCosts[currentState]/
						   quantityInSequentialBatch);
		sequentialBatchWaitingListIndex.next();
	    }

	if(residualTime>1)
	    {
		ourSequentialBatchSpecificationSet
		    .setProductionTimeInTicks(--residualTime);
	    }

	else
	    {
		sequentialBatchWaitingListIndex.setOffset(0);
		for(i=0;i<sequentialBatchWaitingList.getCount();i++)
		    {
			((Order) sequentialBatchWaitingListIndex.get())
			    .setInSequentialBatch(false);
			((Order) sequentialBatchWaitingListIndex.get())
			    .setDoneStep(unitNumber);
			sequentialBatchWaitingListIndex.next();
		    }

		while(sequentialBatchWaitingList.getCount()>0)
		    dailyProductionList.addLast(
			       sequentialBatchWaitingList.removeFirst());

	    }

	    }



    /**
     * production:<br><br>
     *
     * at the beginning of each time tick in the 'day', each unit looks for 
     * the order(s) (if any) to be executed; the production 
     * cycle is 'one unit of time'
     * in a step of the recipe of an order in a tick of a day;<br><br>
     *
     * we can also use inventories
     *
     */
    public void unitStep1 ()
    {
	int i, ii, cycle; 
	boolean lifoShortCut, standAloneBatchProcess;

        // at the beginnig of the tick we have not yet generated variable costs
        totalTickVariableCosts = 0;

	if(StartVEFrame.verbose)
	    System.out.println("Step 1 in unit " + unitNumber +
			       "; waitingList count " + 
			       waitingList.getCount());

	cycle=0;
	//
	// repeating the cycle of unitStep1 if we find zero time orders
	//
	while(cycle==0){
	    cycle++;
	    if(waitingList.getCount()>0)
		if(((Order)waitingList.atOffset(0)).
		   checkZeroTimeProduction())cycle=0;

	// the production
	setUsableWaitingList();
	setStandAloneBatchWaitingList();
	setSequentialBatchWaitingList();

	// after the specific step in recipe production (directly or from 
	// inventories) we place the order(s) in a daily production
	// list, which will be used at unitStep2 to diffuse production
	// results


	if(StartVEFrame.verbose)
	{
	    System.out.println("Stand alone batch steps as seen in unit "
			       +getUnitNumber()+
			       " are "+
			       standAloneBatchWaitingList.getCount()+
			       "; in the assembler are  "+
			       myStandAloneBatchAssembler.
			       getWaitingListLength());
	}

	if(StartVEFrame.verbose)
	{
	    System.out.println("Sequential  batch steps as seen in unit "
			       +getUnitNumber()+
			       " as trapped into the assembler are  "+
			       mySequentialBatchAssembler.
			       getWaitingListLength());
	}

	// managing off line the sequential batches
	if(sequentialBatchWaitingList.getCount()>0)
	    {
		manageSequentialBatchWaitingList();
		return;
	    }


        madeProduction=false;
	if(usableWaitingList.getCount() > 0){
	    firstOrder=(Order)usableWaitingList.removeFirst();
	    waitingList.remove(firstOrder);

	    currentState=-1;
	    for (i=0; i<getNOfPhasesToDealWith() && currentState==-1;i++)
		if(firstOrder.getNextStep()==getProductionPhase(i))
		    currentState=i;
	    if(currentState==-1){
		System.out.println("Unrecoverable error in unit " + unitNumber);
		System.exit(1);
	    }

	    standAloneBatchProcess=false;
	    if(! standAloneBatchWaitingList.contains(firstOrder)){
		dailyProductionList.addLast(firstOrder);
	    }
	    else{
		// decrease counter in proper specificationSet and eliminating
		// the order both from the local stand alone batch list and
		// from the stand alone batch assembler list, if next step
		// is the last one
		if(1==((StandAloneBatchSpecificationSet) 
		 firstOrder.getPendingStandAloneBatchSpecificationSet()).
		   decreaseProductionTime())
		    {
			standAloneBatchWaitingList.remove(firstOrder);
			myStandAloneBatchAssembler.
			    removeThisOrderFromTheWaitingList(firstOrder);
		    }
		standAloneBatchProcess=true;
		usableWaitingList.addFirst(firstOrder);
		waitingList.addFirst(firstOrder);
	    }

//#3
	    if (useWarehouses && getUseWarehouseLocalChoice() 
		&& ! myWarehouse.empty() &&
		getPhaseNumberToProduceInventories()==currentState){
		myWarehouse.decreaseInventoryCounterValue();

//#1 Do we introduce a repetition until inventories are contained in the 
//   warehouse?
//   We have to evaluate the opportunity of doing so; at present the repeated
//   use of inventories arises only if an order has its next step to be done 
//   by the same unit; that order is placed in the order list in a LIFO way 
//   (normally FIFO)

	    }
	    else {
		madeProduction=true;
		addTickVariableCosts();
	    }

	    if(! standAloneBatchProcess)firstOrder.setDoneStep(unitNumber);
	    //set the costs of the order in production

	    firstOrder.setFixedAndVariableOrderCosts(
						     fixedCosts[currentState], 
						     variableCosts[currentState]);

        }

	// having made no production, check to increase inventories
	if(! madeProduction && useWarehouses && getUseWarehouseLocalChoice()){
	    if(useNewses){
		if(inventoryRuleMaster.increaseInventoriesUsingNewses
		   (unitNumber, myWarehouse, newsList)){
		    temporaryCurrentState=currentState;
		    currentState=getPhaseNumberToProduceInventories();
		    myWarehouse.increaseInventoryCounterValue();
		    //recording variable costs
		    addTickVariableCosts();
		    currentState=temporaryCurrentState;
                }
	    }
	    else {
                if(inventoryRuleMaster.increaseInventories
                                          (unitNumber, myWarehouse)){
		    temporaryCurrentState=currentState;
		    currentState=getPhaseNumberToProduceInventories();
		    myWarehouse.increaseInventoryCounterValue();
		    //recording variable costs
		    addTickVariableCosts();
		    currentState=temporaryCurrentState;
                }
	    }
        }

	// the use of inventories
	if (useWarehouses && getUseWarehouseLocalChoice())
	    {
		while(usableWaitingList.getCount()>0 && ! myWarehouse.empty())
		    {
		    myWarehouse.decreaseInventoryCounterValue();
		    firstOrder=(Order)usableWaitingList.removeFirst();
		    waitingList.remove(firstOrder);



	    if(! standAloneBatchWaitingList.contains(firstOrder))
		{
                   firstOrder.setDoneStep(unitNumber);
		   // set the production costs into the order
		   firstOrder.setFixedAndVariableOrderCosts(
					   fixedCosts[currentState], 
					   variableCosts[currentState]);
		    // is the next recipe step to be done in our unit? If yes, 
	            // assign the order immediately to ourselves (if there is 
		    // room in the inventories we can make the production step 
		    // in this same cycle); we manage this step in a LIFO
		    // way to assure continuity in production of a multi-tick
		    // step
		      lifoShortCut=false;
		      for (ii=0; ii<nOfPhasesToDealWith; ii++)
			  if(getProductionPhase(ii)==firstOrder.getNextStep()
			     && getPhaseNumberToProduceInventories()==ii)
			      lifoShortCut=true;
/*waiting*/
		      if(lifoShortCut && cycle !=0) // if(cycle==0) we are 
			                            // executing a zero time
			                            // step; also if the next 
			                            // step has the same code
			                            // it is not the same
			  {
			      waitingList.addFirst(firstOrder);
			      usableWaitingList.addFirst(firstOrder);
			  }
		      // if not, put the order in the daily production list
		      else dailyProductionList.addLast(firstOrder);

		}

	    else
		{
		 // decrease counter in proper specificationSet and eliminating
		 // the order both from the local stand alone batch list and
		 // from the stand alone batch assembler list, if next step
		 // is the last one
		 if(1==((StandAloneBatchSpecificationSet) 
		       firstOrder.getPendingStandAloneBatchSpecificationSet()).
		       decreaseProductionTime())
			{
			    standAloneBatchWaitingList.remove(firstOrder);
			    myStandAloneBatchAssembler.
				removeThisOrderFromTheWaitingList(firstOrder);
			}

		 usableWaitingList.addFirst(firstOrder);
		 waitingList.addFirst(firstOrder);
		}

		   
		    }
	    }

	} //concluding cycle
    }

    /**
     * diffusion and accounting:<br><br>
     *
     * sending the orders to the subsequent production units;<br><br>
     *
     * making accounting operations
     */
    public void unitStep2 () {

	if(StartVEFrame.verbose)
	    System.out.println("Step 2 in unit " + unitNumber +
			       " dailyProductionList count " +
			       dailyProductionList.getCount());

	while(dailyProductionList.getCount()>0)
	    {
		firstOrder=(Order)dailyProductionList.removeFirst();

		// checking for 'or' sequences
		if(firstOrder.checkIfNextStepUsesAnOrSpecificationSet())
		    Or.applyOr(firstOrder);

		// do we have concluded the production step?
		// if yes ... we drop the object
		if (firstOrder.getNextStep()==0)
		    {
			if(StartVEFrame.verbose)
			    System.out.println("drop order # " +
					       firstOrder.getOrderNumber() +
					       " startingTimeUnit " +
					       firstOrder.getStartingTimeUnit() +
					       " endTimeUnit " +
					       Globals.env.getCurrentTime() +
					       " stepNumber " +
					       firstOrder.getStepNumber() +
					       " ratio " +
					       ( (float)
						(Globals.env.getCurrentTime() -
						firstOrder.getStartingTimeUnit()))/
					        firstOrder.getStepNumber());
			// the production is concluded, so we inform
			// our environment (the model) of the time spent
			// and of the # of steps in recipe
			vEFrameModelSwarm.setProductionTimeAndRecipeLength(
			  Globals.env.getCurrentTime() -
			  firstOrder.getStartingTimeUnit(),
			  firstOrder.getStepNumber());
			if(StartVEFrame.verbose)
			    System.out.println("Order #" +
					       firstOrder.orderNumber +
					       " has been completed");
			// remove the dropped order from the wholeOrderList
			wholeOrderList.remove(firstOrder);
                        // recording the revenue from the finished order
                        dailySaleRevenues += firstOrder.getOrderEvaluation();
                        //recording the costs of the order
                        totalFinishedOrderCosts += 
			    firstOrder.getOrderTotalCosts();
                        // now we drop the object (also to save memory space)
			// but first of all we made a log of the 
			// concluded order
			String name=
			    ""+
			    Globals.env.getCurrentTime()/
			    vEFrameModelSwarm.getTicksInATimeUnit()+" "+
			    Globals.env.getCurrentTime()%
			    vEFrameModelSwarm.getTicksInATimeUnit()+" "+
			    firstOrder.getRecipeName()+" "+
			    firstOrder.getOrderNumber()+" "+
			    firstOrder.getStartingTimeUnit()/
			    vEFrameModelSwarm.getTicksInATimeUnit()+" "+
			    firstOrder.getStartingTimeUnit()%
			    vEFrameModelSwarm.getTicksInATimeUnit()+" ("+
			    firstOrder.getStepNumber()+") ";

			for (int m=0; m<firstOrder.getStepNumber();m++)
			    name = name+firstOrder.getOrderRecipe(m)+" ";
			    name = name+"{ ";
			for (int m=0; m<firstOrder.getStepNumber();m++)
			    name = name+firstOrder.getOrderState(m)+" ";
			    name = name+"} ";
			    name = name+
				"["+firstOrder.getOrderTotalCosts()+"]";
			    name = name+" "+firstOrder.getMultiplicity();
			    name = name+"\n";			
			    
			try{
			    concludedOrderLog.write(name);}
			catch(IOException e){
			    System.out.println(e);
			    System.exit(1);
			}

                        firstOrder.drop();
		    }

		// if not ... we send it to the subsequent production unit
		// using unitCriterion (see it in AssigningTool class)
		// if more than one single unit
		// is able to perform the required task
		else assigningTool.assign(firstOrder);
	    }

	// set accounting data
	setTotalDailyCosts();
	if(useWarehouses && getUseWarehouseLocalChoice()){
	    setTotalInventoryFinancialCosts();
	    setStoredComponentDailyValue();
	}

    }

    /*
     * setting daily accounting data (step4 is executed after step3, but
     * it is reported here for practical writing reasons */
    public void unitStep4 ()
    {
	setTotalCosts();
	setTotalSaleRevenues();
    }

    /*
     * dealing with newses
     */
    public void unitStep3 () {

	int n;
	int i, ii, j, offsetValue;
	int m;
        int stepsToBeDone;
	int receivingUnitNumber;
	boolean unitNotFound;

	for(n=0; n<nOfOrdersInNewses && waitingList.getCount()>n && 
		useNewses; n++) 
	    {
		// we get the first order from the waitingList to find how 
		// many steps  are necessary to finish its recipe
		incomingOrder=(Order)waitingList.atOffset(n);
		stepsToBeDone=incomingOrder.getHowManyStepsToBeDone();

		// we ask the order about the subsequent production units, 
		// according information deepness (we start from next step, 
		// in position 2)
		for (i=2;i<=stepsToBeDone && i<=informationDeepness+1;i++)
		    {

			//
			// indentifying the unit to which we can send the news
			//
//#2 Do we have to send newses to all units? and if more than one unit replies?
//   At present we look only for the first one 
			receivingUnitNumber=-1; offsetValue=-1;
			unitNotFound=true;
			for (j=0;j<unitList.getCount() && 
				 unitNotFound;j++)
			    {
				aUnit=(Unit)unitList.atOffset(j);
				for (ii=0;ii<aUnit.getNOfPhasesToDealWith() &&
					 unitNotFound;ii++)
				    {
					if (incomingOrder.
				   getProductionToBeDoneAtStepAheadNumber(i)
					    == aUnit.getProductionPhase(ii))
					    {
						unitNotFound=false;
			 receivingUnitNumber=aUnit.getUnitNumber();
			 offsetValue=j;
					    }
				    }
			    }

			// end unit case (we find the destination unit,
			// without action)
			for (j=0;j<endUnitList.getCount() && 
				 unitNotFound;j++)
			    {
				anEndUnit=(EndUnit) endUnitList.atOffset(j);
				if (incomingOrder.
				    getProductionToBeDoneAtStepAheadNumber(i)
				    == anEndUnit.getUnitNumber())
				    {
					unitNotFound=false;
					receivingUnitNumber=-2;
					offsetValue=-2;
				    }
			    }

			if (receivingUnitNumber==-1 || offsetValue==-1)
			    {
				System.out.println(
 "Internal error attempting to send a news; receiving unit does not exists.");
				System.exit(1);
			    }

			if(!(receivingUnitNumber==-2 && offsetValue==-2))
			    {
				if(StartVEFrame.verbose)
				    System.out.println(
						       " Receiving Unit # " + 
						       receivingUnitNumber);
				//
				// verifying if the informationRuleMaster allows us to
				// send the news and if the wharehouse in the 
				// receiving unit is activated
				//
				if(informationRuleMaster.
				   checkToSendNews(unitNumber, receivingUnitNumber)
				   && unitNumber!= receivingUnitNumber)
				    {
				        // the following is the unit able to make
					// the next production steps; we send the 
					// news if . . .
					receivingUnit=(Unit)unitList.
					    atOffset(offsetValue);
					if (incomingOrder.
					    getProductionToBeDoneAtStepAheadNumber
					    (i) == receivingUnit.
					    getProductionPhase(receivingUnit.
					       getPhaseNumberToProduceInventories()))
					    // if more than one production phase 
					    // can produce 
					    // inventories for its warehouse, 
					    // the first one
					    // is chosed
					    {
					      aNews=new News (getZone(), unitNumber, 
					       receivingUnit.getProductionPhase(
					       receivingUnit.
					       getPhaseNumberToProduceInventories()), 
					       incomingOrder);
					      receivingUnit.setNewsList(aNews);
					    }
				    
				    }
			    }
		    }
	    }
    }

    /** set the necessity of redimensioning the vectors containing the 
     *  parameters; see ($) below */
    public void setRedimensionTrue(){
	nOfPhasesToDealWithDimensionOrRedimension=true;
	nOfCostTypesDimensionOrRedimension=true;
	nOfPhaseInventoriesDimensionOrRedimension=true;
    }


    /** set unit production phases */
    public void setProductionPhase(int ph)
    {
	// ($) the nOfPhasesToDealWith is 1 as default, but it can be 
	// redefined if the unit is a special one (may be, also, newly as 1)
	// to understant how this work, start from VEFrameModelSwarm,
	// line unitParameters.setParametersTo(i, aUnit)
	// 
	// the nOfPhasesToDealWith changes if we introduce a complex unit
	// and so we have to redimension unitProductionPhases vector,
	// zeroing countPhases
	if (nOfPhasesToDealWithDimensionOrRedimension)
	    {
		unitProductionPhases = new int[nOfPhasesToDealWith]; 

		nOfPhasesToDealWithDimensionOrRedimension=false;
		countPhases=0;
	    }
	unitProductionPhases[countPhases] = ph;
	countPhases++;
    }

    /** setting fixed and variable costs */
    public void setFixedAndVariableCosts(float fc, float vc)
    {
	nOfCostTypes=nOfPhasesToDealWith;
	// see ($) above
	if (nOfCostTypesDimensionOrRedimension)
	    {
		fixedCosts    = new float[nOfCostTypes];
		variableCosts = new float[nOfCostTypes];

		nOfCostTypesDimensionOrRedimension=false;
		countCostTypes=0;
	    }
	// our fixed and variable costs
        fixedCosts[countCostTypes] = fc;
        variableCosts[countCostTypes] = vc;
	countCostTypes++;
    }

    /** setting phase inventory choice */
    public void setPhaseInventories(int fi)
    {
	nOfPhaseInventories=nOfPhasesToDealWith;
	// see ($) above
	if (nOfPhaseInventoriesDimensionOrRedimension)
	    {
		phaseInventories    = new boolean[nOfPhaseInventories];

		nOfPhaseInventoriesDimensionOrRedimension=false;
		countPhaseInventories=0;
	    }
	// phase inventory choice
	if (fi==0) phaseInventories[countPhaseInventories] = false;
	else phaseInventories[countPhaseInventories] = true;
	countPhaseInventories++;
    }


    /**
     * return the production phases of the unit
     */
    public int getProductionPhase(int state)
    {
	if (state >=0)
	    return unitProductionPhases[state]; 
	else
	    return -1;
    }

    /**
     * Putting an order in the production waitingList of the unit
     * (using a related objectif any) and deleting the related news
     */
    public void setWaitingList(Order anOrder)
    {
       boolean newsToBeDroppedNotFound;
       int i; 

       // a "p" or "c" or "batch" or "||" or "&&" process

       // case "p": the order is included in myProcurementAssembler
       // waitingList; until it is in this list, it cannot be processed
       // in unitStep1 method
       // if the procurement is immediately completed, the order will be
       // immediately freed in the successive schedule step
       if(anOrder.checkIfNextStepUsesAProcurementSpecificationSet())
	   myProcurementAssembler.setProcurementWaitingList(anOrder);

       // case "stand alone batch": the order is included in 
       // myStandAloneBatchAssembler waitingList; 
       // until it is in this list, it can be processed
       // in unitStep1 method, but it has to be replaced in first waitingList
       // position until myStandAloneBatchAssembler free it, having
       // made the required number of production steps
       if(anOrder.checkIfNextStepUsesAStandAloneBatchSpecificationSet())
	   myStandAloneBatchAssembler.setStandAloneBatchWaitingList(anOrder);

       // case "sequential batch": the order is included in 
       // mySequentialBatchAssembler waitingList; 
       // 
       // order in this list will be cleared for production only when a whole
       // sequential batch exists into the list; the order wil be arranged
       // all after the last of the batch and will be cleared for propagation
       // only after the production of all the batch members
       if(anOrder.checkIfNextStepUsesASequentialBatchSpecificationSet())
	   mySequentialBatchAssembler.setSequentialBatchWaitingList(anOrder);

       if(anOrder.checkZeroTimeProduction())waitingList.addFirst(anOrder);
       else waitingList.addLast(anOrder);

	if(StartVEFrame.verbose)
	    System.out.println("The unit " + unitNumber +
			       " has received the order # "
			       + anOrder.getOrderNumber());

	if (useNewses && newsList.getCount()>0)
	    //looking for a news containing information about
	    // the order just arrived to the unit
	    for (i=1;i<=newsList.getCount();i++)
		{
		    toBeDroppedNews=(News)newsList.removeFirst();
		    if (toBeDroppedNews.incomingOrderInNews.getOrderNumber()
			==anOrder.getOrderNumber())
			{
			    toBeDroppedNews.drop();
			    if(StartVEFrame.verbose)
				System.out.println(
					"Dropping the news announcinig the order n "
					+ toBeDroppedNews.incomingOrderInNews.
					                     getOrderNumber());
			}
		    else
			newsList.addLast(toBeDroppedNews);
		}
    }

    /**
     * return the waiting list length
     */
    public int getWaitingListLength()
    {
	return waitingList.getCount();
    }

    /**
     * removing from the waiting list
     */
    public void removeFromWaitingList(Order o)
    {
	waitingList.remove(o);
    }

    /**
     * putting a news in the newsList of the unit after
     * checking list to avoid duplicated news
     */
    public void setNewsList(News aNews)
    {
	int i;
	boolean  duplicatedNews;
	
	// a single news of the newsList
	News controlNews;
	// the news just arrived
	News toBeCheckedNews;

	toBeCheckedNews=aNews;
	duplicatedNews=false;

	if (newsList.getCount()>0)
	    {
		for (i=0;i<newsList.getCount() && ! duplicatedNews;i++)
		    {
			controlNews=(News) newsList.removeFirst();
			newsList.addLast(controlNews);
			if (toBeCheckedNews.incomingOrderInNews.getOrderNumber()==
			    controlNews.incomingOrderInNews.getOrderNumber()
			    && toBeCheckedNews.getNumberOfTheStepForWhichANewsIsSent()==
			    controlNews.getNumberOfTheStepForWhichANewsIsSent())

			    duplicatedNews=true;
		    }

		if (! duplicatedNews)
		    {
                  newsList.addLast(toBeCheckedNews);
                  if(StartVEFrame.verbose)
		      System.out.println("Unit " + unitNumber +
					 " accept a news containing order # "
					 + toBeCheckedNews.incomingOrderInNews.
					 getOrderNumber());
		    }
		else
		    if(StartVEFrame.verbose)
			System.out.println("Unit " + unitNumber +
					   " doesn't accept a news containing order # "
					   + toBeCheckedNews.incomingOrderInNews.
					   getOrderNumber()+
					   " which just exists for the same step");
            }
	else
	    {
		newsList.addLast(toBeCheckedNews);
		if(StartVEFrame.verbose)
		    System.out.println("Unit " + unitNumber +
				       " accepts a news containing order N "
				       + toBeCheckedNews.incomingOrderInNews.
				       getOrderNumber()+
				       " because newsList is empty");
	    }
    }

    /** set the nOfPhasesToDealWith, i.e. the n. of different steps
     *  the unit is able to deal with */
    public void setNOfPhasesToDealWith(int ns)
    {
	nOfPhasesToDealWith=ns;
    }

    /** get the nOfPhasesToDealWith, i.e. the n. of different steps
     *  the unit is able to deal with */
    public int getNOfPhasesToDealWith()
    {
	return nOfPhasesToDealWith;
    }

    /** set the value of the total daily variable costs */
    public void addTickVariableCosts(){
        totalTickVariableCosts += variableCosts[currentState];
    }

    /** set the global sum of fixed (per day) and variable costs (per tick) 
     *  it can be calculated several times in a day, as a sequence of 
     *  partial values */
    public void setTotalDailyCosts(){
	int i;
	//if currentState=-1 (no production, so no state chosen in the unit)
	if(currentState == -1)
		totalDailyCosts += fixedCosts[0] + totalTickVariableCosts;
	else    totalDailyCosts += 
		    fixedCosts[currentState] + totalTickVariableCosts;
    }

    /** cumulate the total daily costs 
     *  it must be executed only once a day*/
    public void setTotalCosts(){
      totalCosts += totalDailyCosts;
    }

    /** return total daily costs */
    public float getTotalDailyCosts(){
      return totalDailyCosts;
    }

    /** return total costs */
    public float getTotalCosts(){
      return totalCosts;
    }

    /** record the revenue from a dropped order */
    public float getDailySaleRevenues(){
      return dailySaleRevenues;
    }

    /** cumulatere the revenues from finished orders */
    public void setTotalSaleRevenues(){
      totalSaleRevenues += dailySaleRevenues;
    }

    /** return the total revenue */
    public float getTotalSaleRevenues(){
      return totalSaleRevenues;
    }

    /** return total costs from finished orders */
    public float getTotalFinishedOrderCosts(){
	return totalFinishedOrderCosts;
    }

    /** record the cumulated financial costs of inventories */
    public void setTotalInventoryFinancialCosts(){
        totalInventoryFinancialCosts += (myWarehouse.getInventoryQuantity()*
             (vEFrameModelSwarm.getInventoryFinancialRate()/200));
	        // getInventoryFinancialRate returns an yearly rate, which is here
	        // divided by 200 day of labour per year in an approximate way
    }

    /** return the financial costs of inventories */
    public float getTotalInventoryFinancialCosts(){
        return totalInventoryFinancialCosts;
    }

    /** record the daily value of inventories */
    public void setStoredComponentDailyValue(){
	int typeOfInventories;
	// type of inventories, to apply costs
	typeOfInventories=getPhaseNumberToProduceInventories();
        if(vEFrameModelSwarm.getInventoryEvaluationCriterion()==3)
	    dailyStoredComponentValue = myWarehouse.getInventoryQuantity()*
		// to be improved
		vEFrameModelSwarm.getRevenuePerEachRecipeStep();

        else if(vEFrameModelSwarm.getInventoryEvaluationCriterion()==2)
	    dailyStoredComponentValue = myWarehouse.getInventoryQuantity()*
		(fixedCosts[typeOfInventories] + 
		 variableCosts[typeOfInventories]);

        else if(vEFrameModelSwarm.getInventoryEvaluationCriterion()==1)
	    dailyStoredComponentValue = myWarehouse.getInventoryQuantity()*
		variableCosts[typeOfInventories];

        else if(vEFrameModelSwarm.getInventoryEvaluationCriterion()!=1 &&
		vEFrameModelSwarm.getInventoryEvaluationCriterion()!=2 &&
		vEFrameModelSwarm.getInventoryEvaluationCriterion()!=3)
	    {
		System.out.println("your criterion doesn't exist: the variable"+
				   " inventoryEvaluationCriterion must 1, 2 or 3");
		System.exit(1);
	    }
    }

    public float getDailyStoredComponentValue(){
        return dailyStoredComponentValue;
    }

    public boolean getUseWarehouseLocalChoice(){
	boolean test; int i;
	test=false;
	for (i=0;i<nOfPhaseInventories;i++)
	    if (! test) test = phaseInventories[i];
	return test;
    }

    public int getPhaseNumberToProduceInventories(){
	int i, k;
	k=-1;
	for (i=0; i<nOfPhaseInventories && k==-1; i++)
	    if(phaseInventories[i]) k=i;
	return k;
	    }

    /**
     * PTHistogramPlottable interface method: getLabel() 
     */
    public String getLabel(){
	return "" + getUnitNumber();
    }

    /**
     * PTHistogramPlottable interface method: getValueToPlot() 
     */
    public double getValueToPlot(){
	return (double) getWaitingListLength();
    }

}
