// VEFrameModelSwarm.java

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;
import swarm.defobj.SymbolImpl;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.SwarmImpl;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;

import java.io.*;
import java.util.*;

/**
 * The Model contains the Units and all the related tools, like the
 * warehouses which are necessary to deal with the inventories or the
 * newses
 *
 * @author Pietro Terna, with contributions of Marco Remondino, Dario Landini, 
 * Nicola Ormezzano and Paolo Pelligra
 */
public class VEFrameModelSwarm extends SwarmImpl
{
    // simulation parameters

    /** the number of ticks of the clock in a day; at each tick a unit
     * of each step in recipes is done 
     */ 
    public int ticksInATimeUnit;
    /** the total number of operating Units we are using*/
    public int totalUnitNumber;
    /** the total number of end units we are using; and endUnit represents
     *  an actual or virtual warehouse were we can "find" components 
     *  internallly produced using sub-recipes or external procurements; 
     *  each endUnit has a code which is the same used to conclude (as 
     *  an 'end') the recipe describing this internal or external 
     *  production */
    public int totalEndUnitNumber;
    /** the max number of steps to be done to complete an order*/
    public int maxStepNumber;
    /** the max length of a step in a recipe (in units of time or ticks) */
    public int maxStepLength;
    /** the max inventory quantity in a warehouse */
    public int maxInWarehouses;
    /** the min inventory quantity in a warehouse */
    public int minInWarehouses;
    /** choosing if we use or not warehouses (and inventories) */
    public boolean useWarehouses;
    /** choosing if we use or not newses  */
    public boolean useNewses;
    /** how deep information are propagated (how many steps we examine after
     * the current production phase)*/
    public int infDeepness;
    /** the estimate of the revenue of the enterprise per each recipe step 
     * done in a semi manufactured order */
    public float revenuePerEachRecipeStep;
    /** the yearly rate used to evaluate the financial costs of inventories */
     public  float inventoryFinancialRate;

    /** the number of newses necessary to decide the production of 
     *  inventories */
    public int nOfNewsesToProduce;
    /** the number of newses to be cleared after the decision of producing
     *  to increase inventories (a simple choice is to have it equal to 
     *  nOfNewsesToProduce) */
    public int nOfNewsesToBeCleared;
    /** the number of orders for which newses are propagated by a unit */
    public int nOfOrdersInNewses;
    /** the inventory evaluation criterion */
    public  int inventoryEvaluationCriterion;
    /** the total production time used by done orders */
    public int totalProductionTime;
    /** the total length of the recipes of the done orders */
    public int totalRecipeLength;

    /** the order generator, to be replaced by the order distiller or 
     *  in the future by more sophisticated structures (swarms of 
     *  enterprises etc.) */
    public OrderGenerator orderGenerator;

    /** use the order distiller instead of the order generator */
    public boolean useOrderDistiller;

    /** the order distiller, using archives of recipes and of order sequences
     *  to generate the simulated events */
    public OrderDistiller orderDistiller;

    /** ActionGroup for holding an ordered sequence of action */
    public ActionGroup modelActions1, modelActions2, modelActions2b, 
	modelActions2generator, modelActions2distiller, modelActions3;

    /** the Schedule operating in the Model*/
    public Schedule modelSchedule;

    /** the inventoryRuleMaster managing inventories */
    public InventoryRuleMaster inventoryRuleMaster;

    /** the informationRuleMaster managing newses */
    public InformationRuleMaster informationRuleMaster;
    /** the input class for unit parameters*/
    public UnitParameters unitParameters;

    /** the list of the operating units */
    public ListImpl unitList;
    /** its iterator */
    public ListIndex unitListIndex;
    /** the list of the warehouses */
    public ListImpl warehouseList;
    /** the whole list of the orders (not yet finished) */
    public ListImpl orderList;
    /** its iterator */
    public ListIndex orderListIndex;

    /** the list of the end units */
    public ListImpl endUnitList;

    /** the list of procurementAssembler units linked to normal unit; 
     *  proc. ass. units have in charge the task of making the procurement */
    public ListImpl procurementAssemblerList;

    /** the list of standAloneBatchAssembler units linked to normal unit */
    public ListImpl standAloneBatchAssemblerList;

    /** the list of sequentialBatchAssembler units linked to normal unit */
    public ListImpl sequentialBatchAssemblerList;

    /** the log of the concluded orders */
    public MyWriter concludedOrderLog;

    /** the criterion used dealing with 'or' sequences */
    public int orCriterion;

    /** the criterion used to choose the subsequent unit when more the one unit
     *  is able to perform the required task */
    public int unitCriterion;

    /** the tool to assign order to units */
    AssigningTool assigningTool;

    /** the multiplicity used with OrderDistiller, at tick 0 of each
	bunch of ticks */
    public int distillerMultiplicity;

    // the constructor of the model
    public VEFrameModelSwarm (Zone aZone) {
	super (aZone);

	// Now fill in various simulation parameters with default values.
	useOrderDistiller=false;
	ticksInATimeUnit = 1;
	totalUnitNumber = 10;
	totalEndUnitNumber = 3;
	maxStepNumber=5;
	maxStepLength=1;
	useWarehouses=true;
	maxInWarehouses=3;
	minInWarehouses=3;
	useNewses=true;
	infDeepness=3;
	inventoryFinancialRate= (float) 0.1;
	inventoryEvaluationCriterion=1;
	revenuePerEachRecipeStep=2;
	nOfNewsesToProduce=1;
	nOfNewsesToBeCleared=100000;
	nOfOrdersInNewses=1;
	orCriterion=0;
	unitCriterion=0;
	distillerMultiplicity=1;

	// the log of the orders
	try{
	    concludedOrderLog = new MyWriter("log/concludedOrderLog.txt");
	}
	catch (IOException e){
	    System.out.println(e);
	    System.exit(1);
	}
	String text = 
	"Each line contains: final time unit; tick in the final time unit;\n"+
	"                    recipe name; order number\n"+
	"                    starting time unit; tick in the s. time unit\n"+
	"                    (number of steps); the recipe steps\n"+
	"                    { the units that have been doing the various"+
                               " steps of the\n"+
        "                    order (-1=step not executed, 'or' sequence) }\n"+
	"                    [total cost of the order]; multiplicity\n";
	try{
	    concludedOrderLog.write(text);}
	catch(IOException e){
	    System.out.println(e);
	    System.exit(1);
			}

	// Build a customized probe map using a `local' subclass
	// (a special kind of Java `inner class') of the
	// EmptyProbeMapImpl class.
	// This building operation produces a separated class file, with
	// name  VEFrameModelSwarm$1$VEFrameModelProbeMap.class
	class VEFrameModelProbeMap extends EmptyProbeMapImpl {
	    public VarProbe probeVariable (String name) {
		return
		    Globals.env.probeLibrary.getProbeForVariable$inClass
		    (name, VEFrameModelSwarm.this.getClass ());
	    }
	    public MessageProbe probeMessage (String name) {
		return
		    Globals.env.probeLibrary.getProbeForMessage$inClass
		    (name, VEFrameModelSwarm.this.getClass ());
	    }
	    public void addVar (String name) {
		addProbe (probeVariable (name));
	    }
	    public void addMessage (String name) {
		addProbe (probeMessage (name));
	    }
	    public VEFrameModelProbeMap (Zone _aZone, Class aClass) {
		super (_aZone, aClass);
		addVar ("useOrderDistiller");
		addVar ("ticksInATimeUnit");
		addVar ("totalUnitNumber");
	    	addVar ("totalEndUnitNumber");
		addVar ("maxStepNumber");
		addVar ("maxStepLength");
		addVar ("useWarehouses");
		addVar ("useNewses");
		addVar ("maxInWarehouses");
		addVar ("minInWarehouses");
		addVar ("infDeepness");
		addVar ("inventoryFinancialRate");
		addVar ("inventoryEvaluationCriterion");
		addVar ("revenuePerEachRecipeStep");
		addVar ("nOfNewsesToProduce");
		addVar ("nOfNewsesToBeCleared");
		addVar ("nOfOrdersInNewses");
		addVar ("orCriterion");
		addVar ("unitCriterion");
		addVar ("distillerMultiplicity");
    }
	}

	// Install our custom probeMap class directly into the
	// probeLibrary
	Globals.env.probeLibrary.setProbeMap$For
	    (new VEFrameModelProbeMap (aZone, getClass ()), getClass ());
    }
    /** Build the model objects. */
    public Object buildObjects ()
    {
	int i;

	// allow our parent class to build anything
	super.buildObjects();

        // create the class to read acounting parameters
        unitParameters = new 
	    UnitParameters(getZone(), totalUnitNumber, totalEndUnitNumber);
        // read fixed and variable costs
        unitParameters.readUnitData();

	// check maxInWarehouses value and if 0 change it to the max
	// integer value
	if (maxInWarehouses==0)
	    {maxInWarehouses=Integer.MAX_VALUE;
	    // no verbose control here
	    System.out.println(
      "Warning, maxInWarehouses has been changed from 0 to Integer.MAX_VALUE");
	    }

	// the list of the production units
	unitList = new ListImpl (getZone());
	// its iterator
	unitListIndex = unitList.listBegin(getZone());
	// the list of the warehouses
	warehouseList = new ListImpl (getZone());
        // the whole list of the orders existing at each time step
        orderList = new ListImpl(getZone());
	// its iterator
	orderListIndex = orderList.listBegin(getZone());
	// the list of the end units
	endUnitList = new ListImpl (getZone());

	// the instance of assigningTool
	assigningTool = new AssigningTool (getZone(), unitCriterion,
					   unitList, endUnitList);

	// setting 'or' criterion and unitList address in 'Or'
	Or.setOrCriterion(orCriterion);
	Or.setUnitList(unitList);

	// the list of the procurement assemblers linked to the orders 
	// that will be generated 
	procurementAssemblerList = new ListImpl (getZone());

	// the list of the stand alone batch assemblers linked to the orders 
	// that will be generated 
	standAloneBatchAssemblerList = new ListImpl (getZone());

	// the list of the sequential batch assemblers linked to the orders 
	// that will be generated 
	sequentialBatchAssemblerList = new ListImpl (getZone());

	// the orderGenerator or the orderDistiller
	if(! useOrderDistiller)
	    orderGenerator = new OrderGenerator
		(getZone(), maxStepNumber, maxStepLength, unitList, 
		 endUnitList, orderList, this, assigningTool);
	else
	    orderDistiller = new OrderDistiller
		(getZone(), maxStepNumber, maxStepLength, unitList, 
		 endUnitList, orderList, this, assigningTool);


	// the inventoryRuleMaster managing the inventories
	inventoryRuleMaster = new 
	    InventoryRuleMaster(getZone(),
				maxInWarehouses, minInWarehouses,
				nOfNewsesToProduce, nOfNewsesToBeCleared);

        // the informationRuleMaster managing the news propagation
        if(useWarehouses && useNewses)
        informationRuleMaster = new InformationRuleMaster(getZone(),
							  totalUnitNumber);

	// create the endUnits
	for (i = 1; i <= totalEndUnitNumber; i++)
	    {
		EndUnit anEndUnit;

		anEndUnit  =  new EndUnit (getZone(), this);
		unitParameters.setParametersToEndUnit(i, anEndUnit);

		endUnitList.addLast (anEndUnit);
		}

	// create the units and their warehouses, if any
	for (i = 1; i <= totalUnitNumber; i++)
	    {
		Warehouse aWarehouse=null; // null to avoid errors when unused
		News aNews=null;           //  "   "   "     "      "    "
		Unit aUnit;
		ProcurementAssembler aProcurementAssembler;
		StandAloneBatchAssembler aStandAloneBatchAssembler;
		SequentialBatchAssembler aSequentialBatchAssembler;

		// using warehouses and not newses
		if(useWarehouses && ! useNewses)
		{
		    // using the unit number as id also for its warehouse
		    aWarehouse =  new Warehouse (getZone(), unitParameters.
						 getUnitNumber(i));
		    aProcurementAssembler = new ProcurementAssembler 
			(getZone(), endUnitList);
		    aStandAloneBatchAssembler = new 
			StandAloneBatchAssembler (getZone());
		    aSequentialBatchAssembler = new 
			SequentialBatchAssembler (getZone());
		    aUnit  =  new Unit (getZone(), unitList, endUnitList, this,
					inventoryRuleMaster,  aWarehouse,
					orderList, aProcurementAssembler,
					aStandAloneBatchAssembler,
					aSequentialBatchAssembler,
					concludedOrderLog, assigningTool);
		    unitParameters.setParametersTo(i, aUnit);
		    aProcurementAssembler.setUnit(aUnit);
		    aStandAloneBatchAssembler.setUnit(aUnit);
		    aSequentialBatchAssembler.setUnit(aUnit);
		}
		// using both warehoueses and newses
		else if(useWarehouses && useNewses)
		    {
			// using the unit number as id also for its warehouse
			aWarehouse =  new Warehouse (getZone(), unitParameters.
						     getUnitNumber(i));
			aProcurementAssembler = new ProcurementAssembler 
			    (getZone(), endUnitList);
			aStandAloneBatchAssembler = new 
			    StandAloneBatchAssembler (getZone());
			aSequentialBatchAssembler = new 
			    SequentialBatchAssembler (getZone());
			aUnit  =  new Unit(getZone(), unitList, endUnitList,
					   this,
					   inventoryRuleMaster,  aWarehouse,
					   informationRuleMaster, infDeepness, 
				           nOfOrdersInNewses,
				           orderList, aProcurementAssembler,
					   aStandAloneBatchAssembler,
					   aSequentialBatchAssembler,
					   concludedOrderLog, assigningTool);
			unitParameters.setParametersTo(i, aUnit);
			aProcurementAssembler.setUnit(aUnit);
			aStandAloneBatchAssembler.setUnit(aUnit);
			aSequentialBatchAssembler.setUnit(aUnit);
		    }
		// no warehouses and no newses
		else 
		    {
			aProcurementAssembler = new ProcurementAssembler 
			    (getZone(), endUnitList);
			aStandAloneBatchAssembler = new 
			    StandAloneBatchAssembler (getZone());
			aSequentialBatchAssembler = new 
			    SequentialBatchAssembler (getZone());
			aUnit = new Unit  (getZone(), unitList, endUnitList, 
					   this, orderList, 
					   aProcurementAssembler,
					   aStandAloneBatchAssembler,
					   aSequentialBatchAssembler,
					   concludedOrderLog, assigningTool);
			unitParameters.setParametersTo(i, aUnit);
			aProcurementAssembler.setUnit(aUnit);
			aStandAloneBatchAssembler.setUnit(aUnit);
			aSequentialBatchAssembler.setUnit(aUnit);
		    }
		// add the unit, the assembler and the warehouse to the lists
		unitList.addLast (aUnit);
		if(useWarehouses) warehouseList.addLast (aWarehouse);
		procurementAssemblerList.addLast(aProcurementAssembler);
		standAloneBatchAssemblerList.
		    addLast(aStandAloneBatchAssembler);
		sequentialBatchAssemblerList.
		    addLast(aSequentialBatchAssembler);

	    }

	if(! useWarehouses)if(StartVEFrame.verbose)
	    System.out.println("No warehouses generated");


	// creating the dictionary in orderGenerator or in OrderDistiller
	if(! useOrderDistiller)	orderGenerator.setDictionary();
	else                    orderDistiller.setDictionary();

	return this;
    }

    /**
     * Here is where the model schedule is built, the data structures
     * that define the simulation of time in the model. The core is an
     * actionGroup that has a list of actions. That's then put in a
     * Schedule.  */
    public Object buildActions ()
    {
	int i, ii;
	super.buildActions();

	modelActions1 = new ActionGroupImpl (getZone ());
	modelActions2 = new ActionGroupImpl (getZone ());
	modelActions2generator = new ActionGroupImpl (getZone ());
	modelActions2distiller = new ActionGroupImpl (getZone ());
	modelActions2b = new ActionGroupImpl (getZone ());
	modelActions3 = new ActionGroupImpl (getZone ());

	/* zeroing daily accounting variables */
	modelActions1.createActionForEach$message(unitList,
			         SwarmUtils.getSelector("Unit","unitStep0"));
	/* clearing sequential batch production, if any */
	modelActions2.createActionForEach$message(sequentialBatchAssemblerList,
	  SwarmUtils.getSelector("SequentialBatchAssembler","clearToProduce"));
	/* producing and using inventories */
	modelActions2.createActionForEach$message(unitList,
			         SwarmUtils.getSelector("Unit","unitStep1"));
	/* a new order; this step is placed here, after the production step,
	 * to align the diffusion of the order forms (orders under execution 
	 * - next step - or new ones - this step -) */
	if(! useOrderDistiller)
	    modelActions2generator.createActionTo$message(orderGenerator,
			  SwarmUtils.getSelector(orderGenerator,
				 "createRandomOrderWithNSteps"));
	if(useOrderDistiller)
	    modelActions2distiller.createActionTo$message(orderDistiller,
			  SwarmUtils.getSelector(orderDistiller,
			  "distill"));
	/* clearing sequential batch propagation, if any */
	modelActions2b.
	    createActionForEach$message(sequentialBatchAssemblerList,
		      SwarmUtils.getSelector(
			      "SequentialBatchAssembler","clearToPropagate"));
	/* diffusing products (order forms) and making accounting per tick */
	modelActions2b.createActionForEach$message(unitList,
			  SwarmUtils.getSelector("Unit","unitStep2"));
	/* freeing the orders filled with the procurements:
	 * these are orders generated or progagated both immediately above 
	 * and in previous generation and propagation steps */
	modelActions2b.createActionForEach$message(procurementAssemblerList,
		          SwarmUtils.getSelector("ProcurementAssembler",
				     "checkingProcurementsAndFreeingOrders"));
	/* checking the orders processed in stand alone batches */
	modelActions2b.createActionForEach$message(
		       standAloneBatchAssemblerList,
		          SwarmUtils.getSelector("StandAloneBatchAssembler",
				   "checkingStandAloneBatchAndFreeingOrders"));
	/* checking the orders processed in sequential batches */
	modelActions2b.createActionForEach$message(
                       sequentialBatchAssemblerList,
		          SwarmUtils.getSelector("SequentialBatchAssembler",
				   "checkingSequentialBatchAndFreeingOrders"));
	/* propagating newses */
	modelActions2b.createActionForEach$message(unitList,
                          SwarmUtils.getSelector("Unit","unitStep3"));
	/* making daily accounting */
	modelActions3.createActionForEach$message(unitList,
			  SwarmUtils.getSelector("Unit","unitStep4"));

	// Then we create a schedule that executes the
	// modelActions. modelActions is an ActionGroup, by itself it
	// has no notion of time. In order to have it executed in
	// time, we create a Schedule that says to use the
	// modelActions ActionGroup at particular times
	modelSchedule = new ScheduleImpl (getZone (), ticksInATimeUnit);

	modelSchedule.at$createAction (0, modelActions1);

	for (i=0;i<ticksInATimeUnit;i++)
	    {
		modelSchedule.at$createAction (i, modelActions2);
		if(! useOrderDistiller)modelSchedule.
				 at$createAction (i, modelActions2generator);
		if(  useOrderDistiller && i==0)
		    for(ii=0;ii<distillerMultiplicity;ii++)
			modelSchedule.
			     at$createAction (i, modelActions2distiller);
		modelSchedule.at$createAction (i, modelActions2b);
	    }
	modelSchedule.at$createAction (ticksInATimeUnit-1, modelActions3);
	
	return this;
    }

    /**
     * Now set up the model's activation. swarmContext indicates where
     * we're being started in - typically, this model is run as a
     * subswarm of an observer swarm. */
    public Activity activateIn (Swarm swarmContext) {
	// First, activate ourselves via the superclass
	// activateIn: method.  Just pass along the context: the
	// activity library does the right thing.
	super.activateIn (swarmContext);

	// Now activate our own schedule.
	modelSchedule.activateIn (this);

	// Finally, return our activity.
	return getActivity ();
  }
    
    /** the method returns the list of the units */
    public Object getUnitList(){
	return unitList;
    }

    /** the method returns the list of the warehouses */
    public Object getWarehouseList (){
	return warehouseList;
    }

    /** the method returns the list of the orders */
    public Object getOrderList(){
        return orderList;
    }

    /** this get method returns true if warehouses are used, false otherwise */
    public boolean getUseWarehouses() {
	return useWarehouses;
    }

    /** record total production time and total recipe length */
    public void setProductionTimeAndRecipeLength(int pt, int rl) {
	totalProductionTime += pt;
	totalRecipeLength   += rl;
	if(StartVEFrame.verbose)
	    System.out.println("totalProductionTime " + totalProductionTime +
			       " totalRecipeLength " + totalRecipeLength +
			       " ratio " + ((float)
					    totalProductionTime)/totalRecipeLength);
	return;
    }

    /** report the ratio totalProductionTime/totalRecipeLength */
    public float getTimeLengthRatio() {
	if(totalRecipeLength != 0)
	    return ((float) totalProductionTime)/totalRecipeLength;
	else return (float) 1; // 1 is a good choice, being 0 totalRecipeLength
	                       // in the first simulation days and being 1
	                       // the actual ratio of the first
	                       // concluded orders
    }

    /** the financial rate */
    public float getInventoryFinancialRate(){
        return inventoryFinancialRate;
    }

    /** the evaluation criterion for inventories */
    public float getInventoryEvaluationCriterion(){
        return inventoryEvaluationCriterion;
    }

   /** the revenue for each recipe step */
    public float getRevenuePerEachRecipeStep(){
        return revenuePerEachRecipeStep;
    }

    /** the number of ticks in a day */
    public int getTicksInATimeUnit(){
	return ticksInATimeUnit;
    }

    /** the benefit of the enterprise */
    public float getBenefit() {
	float benefit;
        float totalRevenues=0;
        float totalCosts=0;
        float totalSemimanufacturedRevenues=0;
        float dailyStoredComponentRevenues=0;
        float totalInventoryFinCosts=0;
        Unit aUnit;
        Order anOrder;
	int i;

	if(unitList.getCount()>0){
	    unitListIndex.setOffset(0);
	    for(i=0; i<unitList.getCount(); i++){
		aUnit=(Unit)unitListIndex.get();
		totalRevenues += aUnit.getTotalSaleRevenues();
		totalCosts += aUnit.getTotalCosts();
		if(useWarehouses){
		    dailyStoredComponentRevenues += 
			aUnit.getDailyStoredComponentValue();
		    totalInventoryFinCosts += 
			aUnit.getTotalInventoryFinancialCosts();
		}
		unitListIndex.next();
	    }
	}

	if(orderList.getCount()>0){
	    orderListIndex.setOffset(0);
	    for(i=0; i<orderList.getCount(); i++){
		anOrder = (Order)orderListIndex.get();
		totalSemimanufacturedRevenues += anOrder.getOrderEvaluation();
		orderListIndex.next();
	    }
	}

        benefit = totalRevenues + dailyStoredComponentRevenues +
	    totalSemimanufacturedRevenues - totalCosts - 
	    totalInventoryFinCosts;

        return benefit;
    }


}


